if tr and tr.exit then
    tr.exit()
end

-- required by mudlet
mudlet = mudlet or {}; mudlet.mapper_script = true

tr = {
    ui = {}
}

tr.settings = {
    name = "TR", -- information required by logger: name, debug_levels
    debug_levels = {  -- 0 - print no messages (but critical errors), 10 - print everything
    other   = 10 },
    move_delay = 2000, -- time move should take, in ms
}

tr.state = {
    event_handlers = {},
    past = { -- where we were in past move - previous known gmcp frame + direction (if available)
        room_id = nil,
        room_hash = nil,
        area_id = nil,
        area_name = nil,
    },
    present = { -- where are we now (as of received info) - last known gmcp frame
        room_id = nil,
        room_hash = nil,
        area_id = nil,
        area_name = nil,
    },
    mapper = {
        enabled = false,
        mode = "complex", -- simple = link one way, complex = link both ways, smart = link neighbours
        placement_hint = nil, -- for special exits - hint where to place them (x,y,z)
    },
    move_ongoing = false, -- set if at least single move was sent, hit twice disables mapper
    last_dir = nil,
}

local settings = tr.settings
local state = tr.state

--------------------------------------------------------------------------------
-- TODO:
-- position predictor
function tr.moveOverride(dir, forced, speedwalk)

    if state.mapper.enabled and state.move_ongoing then
        log.info(tr, nil, "Moved twice while mapper enabled. Disabling mapper.\n")
        tr.mapperDisable()
    end

    if forced then
        log.info(tr, nil, "Forced Move: %s\n", dir)
        mud.sendCmd(dir, silent)
        return
    end

    -- proces human overrides/blocks only if room is precise
    -- TODO: add predictor from old mapper code
    if not speedwalk and not state.move_ongoing then
        local id = state.present.room_id
        -- check for blocks (then cancel path)
        if id then
            if getRoomUserData(id, "block_"..dir) ~= "" then
                log.err(tr, "walker", "Exit %s blocked: %s\n", dir, getRoomUserData(id, "block_"..dir))
                return
            end
            -- check for overrides (and apply)
            if getRoomUserData(id, "override_"..dir) ~= "" then
                log.info(tr, "walker", "Overriding %s -> %s.\n", dir, getRoomUserData(id, "override_"..dir))
                dir = getRoomUserData(id, "override_"..dir)
            end
        end
    end

    state.move_ongoing = true
    state.last_dir = dir

    log.info(tr, nil, "Move: %s\n", dir)
    mud.sendCmd(dir, silent)
end

function tr.onMovementTimerTimeout()
    state.movement_timer = nil
    raiseEvent("tr_move_timeout")
end

function tr.startMovementTimer()
    if state.movement_timer then
        killTimer(state.movement_timer)
    end
    state.movement_timer = tempTimer(settings.move_delay / 1000, [[tr.onMovementTimerTimeout()]])
end

function tr.eventMovement(event)
    --log.info(tr, nil, "Traveler Event: %s\n", event)
    if event == "obs_move_failed" or event == "obs_move_no_exit" then
        state.move_ongoing = false
        state.mapper.placement_hint = nil
        state.last_dir = nil
    end
    tr.mapperMoved()
    if state.present.room_id then
        centerview(state.present.room_id)
    end
end

-- executes speedWalkPath/...
-- TODO: sync movement
function tr.executePath(run)
    local cur_room = state.present.room_id

    tr.mapperDisable() -- just incase :)

        -- TODO: that is trivial, non-sync speedwalk
    for i=1,#speedWalkDir do
        local next_dir = speedWalkDir[i]
        local next_room = speedWalkPath[i]

        -- check for blocks (then cancel path)
        if getRoomUserData(cur_room, "block_"..next_dir) ~= "" then
            log.err(tr, "walker", "Can't move to %s. That entry is blocked for speedwalk. Reason: %s\n",
                    getRoomName(next_room),  getRoomUserData(cur_room, "block_"..next_dir))
            break
        end
        -- check for overrides (and apply)
        if getRoomUserData(cur_room, "override_"..next_dir) ~= "" then
            log.info(tr, "walker", "Overriding %s -> %s.\n", next_dir,
                     mud.sendMove(getRoomUserData(cur_room, "override_"..next_dir)))
            mud.sendMove(getRoomUserData(cur_room, "override_"..next_dir))
        else -- no overrides - just move
            mud.sendMove(speedWalkDir[i])
        end
        cur_room = next_room
    end
end

function doSpeedWalk()
    if #speedWalkDir == 0 or #speedWalkPath == 0 then
        log.info(tr, "walker", "Speedwalk: no need to move.\n")
        return
    end

    tr.executePath(true)
end


-- list all poi's in current domain (or all domains)
-- TODO: cache poi's
function tr.showPoi(all)
    local pois
    if all then
        pois = searchRoomUserData("poi")
        cecho("All available PoIs:\n")
    else
        if not state.present.area_id or getRoomAreaName(state.present.area_id) == -1 then
            log.err(tr, nil, "Unknown area. Can't dump PoIs.\n")
            return
        end
        cecho("PoIs in <green>"..getRoomAreaName(state.present.area_id)..":\n")
        local rooms = getAreaRooms(state.present.area_id)
        pois = {}
        for k,v in pairs(rooms) do
            local ud = getRoomUserData(v, "poi")
            if ud and ud ~= "" then
                pois[ud] = ud
            end
        end
    end
    for k,v in pairs(pois) do
        cecho("    <green>"..v.."\n")
    end
end

-- human-requested path
-- TODO: multiple rooms with the same poi/description
-- fuzzy: don't match full room name
function tr.travelTo(where, run, fuzzy)
    local pathSet = false
    run = run or false
    fuzzy = fuzzy or false

    if not state.present.room_id then
        log.err(tr, "walker", "Can't go from unknown room.\n")
        return
    end

    -- goto room number - direct
    if tonumber(where) then
        pathSet = getPath(state.present.room_id, tonumber(where))
        if pathSet then
            log.info(tr, "walker", "Going to room # %d - %s\n", where, getRoomName(tonumber(where)))
        end
    end

    -- goto poi (current domain)
    -- try finding nearest room
    if not pathSet and state.present.area_id then
        local rooms = getAreaRooms(state.present.area_id)
        local closest_room = nil
        local closest_cost = nil

        for k,id in pairs(rooms) do
            local ud = getRoomUserData(id, "poi")

            if ud and ud == where then
                local cost
                pathSet, cost = getPath(state.present.room_id, id)
                if pathSet and (not closest_room or cost < closest_cost) then
                    closest_room = id
                    closest_cost = cost
                end
            end
        end
        if pathSet then
            log.info(tr, "walker", "Going to room # %d cost %d - %s\n", closest_room, closest_cost, getRoomName(closest_room))
            pathSet = getPath(state.present.room_id, closest_room)
        end
    end

    -- goto poi (global)
    -- do not bother searching nearest room
    if not pathSet then
        local rooms = searchRoomUserData("poi", where)

        for idx, id in pairs(rooms) do
            pathSet = getPath(state.present.room_id, id)
            if pathSet then
                log.info(tr, "walker", "Going to room # %d in %s - %s\n", id, getRoomAreaName(getRoomArea(id)), getRoomName(id))
                break
            end
        end
    end

    -- goto room name
    if not pathSet then
        local rooms = searchRoom(where, not fuzzy, not fuzzy)
        local closest_room = nil
        local closest_distance = nil

        -- zero: maybe we're already there ;)
        if rooms[state.present.room_id] then
            log.info(tr, "walker", "No need to go to named room - we are already here: %s.\n", tostring(rooms[state.present.room_id]))
            return
        end
        -- first try to find room in current area
        -- posibly closest one
        for id, name in pairs(rooms) do
            if (getRoomArea(id) == state.present.area_id) then
                local cost
                pathSet, cost = getPath(state.present.room_id, id)
                if pathSet and (not closest_room or cost < closest_cost) then
                    closest_room = id
                    closest_cost = cost
                end
            end
        end
        if pathSet then
            log.info(tr, "walker", "Going to room %d (by name) cost %d - %s\n", closest_room, closest_cost, getRoomName(closest_room))
            pathSet = getPath(state.present.room_id, closest_room)
        end

        -- then find everywhere
        if not pathSet then
            for id, name in pairs(rooms) do
                pathSet = getPath(state.present.room_id, id)
                if pathSet then
                    log.info(tr, "walker", "Going to room %d in %s (by name) - %s\n", id, getRoomAreaName(id), name)
                    break
                end
            end
        end
    end

    if not pathSet then
        log.err(tr, "walker", "Unable to find path.\n")
        return
    end

    tr.executePath(run)
end


function tr.eventGmcpRoom()
    tr.state.gmcp_room = utils.cloneTable(gmcp.room.info)

    local s = state
    local room_hash = tr.state.gmcp_room.id
    -- TODO: should we handle empty room id?
    -- todo : That is not 100% reliable, but worst we can risk is messing up new map
    s.move_ongoing = false

    if s.present.room_hash ~= room_hash then
        s.past = utils.cloneTable(s.present)
        s.present.room_hash = room_hash
        s.present.room_id = getRoomIDbyHash(s.present.room_hash)
        if s.present.room_id == -1 then
            s.present.room_id = nil
            s.present.area_id = nil
            s.present.area_name = nil
        else
            s.present.area_id = getRoomArea(s.present.room_id)
            s.present.area_name = getRoomAreaName(s.present.room_id)
        end
        tr.updateRoomInfo()
        raiseEvent("tr_moved")
    end
end

function tr.briefExitColor(room_id, dir)
    local short = tr.dirToShort(dir)
    if short then
        if getRoomUserData(room_id, "block_"..short) ~= "" then
            return "<red>"
        end

        if getRoomUserData(room_id, "override_"..short) ~= "" then
            return "<cyan>"
        end
    end

    return "<green>"
end

function tr.init()
    log.info(tr, nil, "Remapping movement.\n")
    mud.sendMoveOverride = tr.moveOverride
    tr.initEvents()

    -- create room infopanel
    tr.ui.room_info = Geyser.MiniConsole:new({name="tr.infopanel", x=5, y=5, width=-5, height=-5}, ui.extra:getLayout())
    ui.extra:add("map", tr.ui.room_info)
    tr.ui.room_info:setFontSize(10)
    setBackgroundColor(tr.ui.room_info.name, 10, 10, 10, 100)
    setTextFormat(tr.ui.room_info.name, 10, 10, 10, 255, 255, 255, false, false, false)
    setWindowWrap(tr.ui.room_info.name, 95)
    tr.ui.room_info:clear()

    if gmcp and gmcp.room and gmcp.room.info then
        -- fake move on reload to setup map location
        tr.eventGmcpRoom("gmcp.room.info")
    end
end

function tr.exit()
    mud.sendMoveOverride = nil
    tr.exitEvents()
    if state.movement_timer ~= nil then
        killTimer(state.movement_timer)
    end
end

function tr.initEvents()
    log.info(tr, nil, "Initializing event handlers...\n")
    state.event_handlers.gmcp_room = registerAnonymousEventHandler("gmcp.room.info",   tr.eventGmcpRoom)
    state.event_handlers.move_failed = registerAnonymousEventHandler("obs_move_failed",   tr.eventMovement)
    state.event_handlers.move_no_exit = registerAnonymousEventHandler("obs_move_no_exit", tr.eventMovement)
    state.event_handlers.move_timeout = registerAnonymousEventHandler("tr_move_timeout", tr.eventMovement)
    state.event_handlers.movement = registerAnonymousEventHandler("tr_moved", tr.eventMovement)
end

function tr.exitEvents()
    log.info(tr, nil, "Clearing event handlers...\n")
    for k,v in pairs(state.event_handlers) do
        killAnonymousEventHandler(v)
    end

    state.event_handlers = {}
end

-- Initialize
if tr and tr.init then
    genLoad("traveler-room")
    genLoad("traveler-mapper")
    genLoad("traveler-config")
    genLoad("traveler-user", true)
    tr.init()
end
